#region usings
using System;
using System.ComponentModel.Composition;
using System.Runtime.InteropServices;

using SlimDX;
using SlimDX.Direct3D9;
using VVVV.Core.Logging;
using VVVV.PluginInterfaces.V1;
using VVVV.PluginInterfaces.V2;
using VVVV.PluginInterfaces.V2.EX9;
using VVVV.Utils.VColor;
using VVVV.Utils.VMath;
using VVVV.Utils.SlimDX;

#endregion usings

//here you can change the vertex type
using VertexType = VVVV.Utils.SlimDX.TexturedVertex;

namespace VVVV.Nodes
{
	#region PluginInfo
	[PluginInfo(Name = "TextureBlendTest", Category = "EX9.Texture", Version = "0.1", Help = "Basic template which creates a texture", Tags = "")]
	#endregion PluginInfo
	public class C0_1EX9_TextureTextureBlendTestNode : DXTextureOutPluginBase, IPluginEvaluate
	{
		#region fields & pins

		[Input("Wave Count", DefaultValue = 1)]
		IDiffSpread<double> FWaveCountIn;

		[Input("Width", DefaultValue = 64)]
		IDiffSpread<int> FWidthIn;

		[Input("Height", DefaultValue = 64)]
		IDiffSpread<int> FHeightIn;
		
		[Input("Enable Blend", DefaultValue = 0, IsSingle = true)]
		IDiffSpread<bool> FBlendEnableIn;

		[Input("Blend", DefaultValue = 0.5, IsSingle = true)]
		IDiffSpread<double> FBlendIn;

		[Input("TextureU Change", DefaultValue = 0.0, IsSingle = true)]
		IDiffSpread<float> FTextureUChangeIn;

		[Input("TextureV Change", DefaultValue = 0.0, IsSingle = true)]
		IDiffSpread<float> FTextureVChangeIn;

		[Input("Switch Textures", DefaultValue = 0, IsSingle = true)]
		IDiffSpread<bool> FSwitchTexturesIn;


		[Import()]
		ILogger FLogger;

		//track the current texture slice
		int FCurrentSlice;

		#endregion fields & pins

		#region irrelevant
		
		// import host and hand it to base constructor
		[ImportingConstructor()]
		public C0_1EX9_TextureTextureBlendTestNode(IPluginHost host) : base(host)
		{
		}

		//called when data for any output pin is requested
		public void Evaluate(int SpreadMax)
		{
			SetSliceCount(2);

			//recreate texture if resolution was changed
			if (FWidthIn.IsChanged || FHeightIn.IsChanged) {
				//set new texture size
				Reinitialize();
			}

			//update texture
			if (FWaveCountIn.IsChanged) {
				Update();
			}

			//update texture
			if (	FBlendIn.IsChanged || FBlendEnableIn.IsChanged 
				 || FTextureUChangeIn.IsChanged || FTextureVChangeIn.IsChanged
				 || FSwitchTexturesIn.IsChanged ) {
				Update();
			}
			
		}

		//this method gets called, when Reinitialize() was called in evaluate,
		//or a graphics device asks for its data
		protected override Texture CreateTexture(int Slice, Device device)
		{
			FLogger.Log(LogType.Debug, "Creating new texture at slice: " + Slice);
			//return TextureUtils.CreateTexture(device, Math.Max(FWidthIn[Slice], 1), Math.Max(FHeightIn[Slice], 1));
			Texture t;
			if (Slice == 0) {
				t = TextureUtils.CreateTexture(device, Math.Max(FWidthIn[Slice], 1), Math.Max(FHeightIn[Slice], 1)); 
			}
			else {
				try {
					t = new Texture(device, Math.Max(FWidthIn[Slice], 1), Math.Max(FHeightIn[Slice], 1), 1, Usage.RenderTarget, Format.A8R8G8B8, Pool.Default);
				} catch (Exception e) {
					//device.Capabilities
					FLogger.Log(LogType.Error, e.Message);
					t = TextureUtils.CreateTexture(device, Math.Max(FWidthIn[Slice], 1), Math.Max(FHeightIn[Slice], 1)); 
					//  new Texture(device, Math.Max(FWidthIn[Slice], 1), Math.Max(FHeightIn[Slice], 1), 1, Usage.None, Format.A8B8G8R8, Pool.Managed);
				}
				//SlimDX.Direct3D9.Surface.CreateRenderTarget(device,Math.Max(FWidthIn[Slice], 1), Math.Max(FHeightIn[Slice], 1), Format.A8B8G8R8, MultisampleType.None, 0, true);
			}
			return t;
			
		}

		#endregion irrelevant

		//this method gets called, when Update() was called in evaluate,
		//or a graphics device asks for its texture, here you fill the texture with the actual data
		//this is called for each renderer, careful here with multiscreen setups, in that case
		//calculate the pixels in evaluate and just copy the data to the device texture here
		unsafe protected override void UpdateTexture(int Slice, Texture texture)
		{
			if (Slice == 0) {
				if ( FSwitchTexturesIn[0] ) {
					TextureUtils.Fill32BitTexInPlace(texture, FillTexure2 );
				}
				else {
					TextureUtils.Fill32BitTexInPlace(texture, FillTexure1 );
				}
			}
			else {
				FCurrentSlice = Slice;
				//TextureUtils.Fill32BitTexInPlace(texture, FillTexure);
				
				try {
					Device d = texture.Device;
					Texture tA = TextureUtils.CreateTexture(texture.Device, Math.Max(FWidthIn[Slice], 1), Math.Max(FHeightIn[Slice], 1));
					TextureUtils.Fill32BitTexInPlace(tA, FillTexure1);
					Texture tB = TextureUtils.CreateTexture(texture.Device, Math.Max(FWidthIn[Slice], 1), Math.Max(FHeightIn[Slice], 1));
					TextureUtils.Fill32BitTexInPlace(tB, FillTexure2);
					
					d.BeginScene();
					d.SetRenderTarget(0, texture.GetSurfaceLevel(0));
					
					//this way we see where the texture is
					d.ColorFill(texture.GetSurfaceLevel(0), new Color4( 1.0f, 0.1f, 0.1f, 0.1f ) );
					
					d.SetTransform( TransformState.View, Matrix.LookAtLH( new Vector3(0, 0, -2),  new Vector3(0, 0, 0), new Vector3(0, 1, 0) ) );
					d.SetTransform( TransformState.Projection, Matrix.OrthoLH(2, 2, 1, 1000) );
					d.SetTransform( TransformState.World, Matrix.Identity );
					//d.Clear(ClearFlags.All, new Color4(), 10.0f, 0);
					d.SetTransform( TransformState.Texture0, Matrix.Identity );
					d.SetTransform( TransformState.Texture1, Matrix.Identity );
					//d.SetTransform( TransformState., Matrix.Identity );
					
					d.SetRenderState(RenderState.Lighting, false);
					d.SetRenderState(RenderState.CullMode, Cull.None);
					d.SetRenderState(RenderState.ZEnable, false);
		
					d.SetSamplerState(0, SamplerState.AddressU, TextureAddress.Wrap);
					d.SetSamplerState(0, SamplerState.AddressV, TextureAddress.Wrap);
					d.SetSamplerState(0, SamplerState.AddressW, TextureAddress.Wrap);
	
					d.SetTextureStageState(0, TextureStage.TexCoordIndex, 0);
					d.SetTextureStageState(0, TextureStage.TextureTransformFlags, TextureTransform.Count3);

					//d.SetTextureStageState(0, TextureStage.TextureTransformFlags, TextureTransform.Disable);
					d.SetTextureStageState(0, TextureStage.ColorArg1, TextureArgument.Texture);
					d.SetTextureStageState(0, TextureStage.ColorOperation, TextureOperation.SelectArg1);
					d.SetTextureStageState(0, TextureStage.AlphaArg1, TextureArgument.Texture);
					d.SetTextureStageState(0, TextureStage.AlphaOperation, TextureOperation.SelectArg1);
	
					d.SetTexture(0, (FSwitchTexturesIn[0] ? tB : tA) );
					
					if (FBlendEnableIn[0]) {
						int blend = (int)(Math.Max(0.0f, Math.Min(FBlendIn[Slice], 1.0f)) * 255);
						int blendRGBA = blend * 0x1000000 + blend * 0x10000 + blend * 0x100 + blend;
						d.SetRenderState(RenderState.TextureFactor, blendRGBA );
						//d.SetRenderState(RenderState.BlendFactor, blendRGBA );
						//d.SetRenderState(RenderState.BlendOperationAlpha, BlendOperation.Minimum);
						//d.SetRenderState(RenderState.AlphaBlendEnable, true);
						//d.SetRenderState(RenderState.Wrap0, TextureWrapping.None);
						//d.SetRenderState(RenderState.Wrap1, TextureWrapping.None);
						//d.SetRenderState(RenderState.Wrap2, TextureWrapping.None);
						//d.SetRenderState(RenderState.Wrap3, TextureWrapping.None);
	
						d.SetTextureStageState(1, TextureStage.TextureTransformFlags, TextureTransform.Disable);
						d.SetSamplerState(1, SamplerState.AddressU, TextureAddress.Wrap);
						d.SetSamplerState(1, SamplerState.AddressV, TextureAddress.Wrap);
						d.SetSamplerState(1, SamplerState.AddressW, TextureAddress.Wrap);
						
						d.SetTextureStageState(1,TextureStage.ColorArg1, TextureArgument.Texture); //arg 1 is texture
						d.SetTextureStageState(1, TextureStage.ColorArg2, TextureArgument.Current); //arg 2 is last stage
						d.SetTextureStageState(1, TextureStage.ColorOperation, TextureOperation.BlendFactorAlpha); //TextureOperation.BlendFactorAlpha);

						d.SetTextureStageState(1,TextureStage.AlphaArg1, TextureArgument.Texture); //arg 1 is texture
						d.SetTextureStageState(1,TextureStage.AlphaArg2, TextureArgument.Current); //arg 1 is texture
						d.SetTextureStageState(1, TextureStage.AlphaOperation, TextureOperation.BlendFactorAlpha);
	
						d.SetTexture(1, (FSwitchTexturesIn[0] ? tA : tB) );
	
					}
					
					
	
					VertexBuffer vb = CreateQuadTexturedVertexBuffer(d);
					
					d.SetStreamSource(0, vb, 0, Marshal.SizeOf( typeof( TexturedVertex ) ) );
					if (FBlendEnableIn[0]) {
						d.SetStreamSource(1, vb, 0, Marshal.SizeOf( typeof( TexturedVertex ) ) );
						
						//or copy vertexbuffer from 1st stage
						//int vbOffset;
						//int vbStride;
						//d.GetStreamSource(0, out vb, out vbOffset, out vbStride);				
						//d.SetStreamSource(1, vb, vbOffset, vbStride );
					}
					
					
					
					d.DrawPrimitives(PrimitiveType.TriangleStrip, 0, 2);
	
	
					DataStream ds = vb.Lock(0, 0, LockFlags.ReadOnly);
					string s = "datastream: ";
					try {
						while (ds.Position < ds.Length) {
								s += " | " + ds.Read<float>();
						}
					} catch {}
					FLogger.Log(LogType.Debug, s);
					vb.Unlock();
					
					if ( d.Capabilities.MaxSimultaneousTextures <= 1) {
						FLogger.Log(LogType.Debug, "DRAWING will fail");
					}
					else {
						FLogger.Log(LogType.Debug, "DRAWING (up to " + d.Capabilities.MaxSimultaneousTextures + " simultaneous textures) " + "4 x vertex size = " + Marshal.SizeOf( typeof( TexturedVertex ) )
									+ " format:" + vb.Description.Format.ToString() + " fvf:" + vb.Description.FVF.ToString() + " pool:" + vb.Description.Pool.ToString() + " size:" + vb.Description.SizeInBytes.ToString() + " usage:" + vb.Description.Usage.ToString() 
									//+ "\nAddresMode: U" + d.GetSamplerState(0, SamplerState.AddressU) + " V" + d.GetSamplerState(0, SamplerState.AddressV) + " W" + d.GetSamplerState(0, SamplerState.AddressW)
						);
					}
					
					
					d.EndScene();
	
					d.SetTextureStageState(1, TextureStage.ColorOperation, TextureOperation.Disable);
	
				} catch (Exception e) {
					FLogger.Log(LogType.Error, "TEST PROBLEM: " + e.Message);
				}
			}
		}

		#region FillTexture functions
		
		//this is a pixelshader like method, which we pass to the fill function
		unsafe private void FillTexure(uint* data, int row, int col, int width, int height)
		{
			//crate position in texture in the range [0..1]
			var x = (double)row / height;
			var y = (double)col / width;

			//make some waves in the range [0..255]
			var wave = (byte)(255 * ((Math.Sin(x * y * FWaveCountIn[FCurrentSlice] * 10) + 1) / 2.0));

			//a pixel is just a 32-bit unsigned int value
			var pixel = UInt32Utils.fromARGB(wave, wave, wave, wave);

			//copy pixel into texture
			TextureUtils.SetPtrVal2D(data, pixel, row, col, width);
		}

		unsafe private void FillTexureBlue(uint* data, int row, int col, int width, int height)
		{
			//crate position in texture in the range [0..1]
			var x = (double)row / height;
			var y = (double)col / width;

			//make some waves in the range [0..255]
			var wave = (byte)(255 * ((Math.Sin(x * y * FWaveCountIn[FCurrentSlice] * 10) + 1) / 2.0));

			//a pixel is just a 32-bit unsigned int value
			var pixel = UInt32Utils.fromARGB(wave, 0, 0, wave);

			//copy pixel into texture
			TextureUtils.SetPtrVal2D(data, pixel, row, col, width);
		}

		//this is a pixelshader like method, which we pass to the fill function
		unsafe private void FillTexureRed(uint* data, int row, int col, int width, int height)
		{
			//crate position in texture in the range [0..1]
			var x = 1 - (double)row / height;
			var y = (double)col / width;

			//make some waves in the range [0..255]
			var wave = (byte)(255 * ((Math.Sin(x * y * FWaveCountIn[FCurrentSlice] * 20) + 1) / 2.0));

			//a pixel is just a 32-bit unsigned int value
			var pixel = UInt32Utils.fromARGB(wave, wave, (byte)(255 - wave), 0);

			//copy pixel into texture
			TextureUtils.SetPtrVal2D(data, pixel, row, col, width);
		}

		//horizontal bars, yellow to blue
		unsafe private void FillTexure1(uint* data, int row, int col, int width, int height)
		{
			//create position in texture in the range [0..1]
			var x = (double)row / height;
			var y = 1; //(double)col / width;

			//make some waves in the range [0..255]
			var wave = (byte)(255 * ((Math.Sin(x * y * FWaveCountIn[FCurrentSlice] * 20) + 1) / 2.0));

			//a pixel is just a 32-bit unsigned int value
			byte transparency = wave; //(byte)(wave * (1-x));
			var pixel = UInt32Utils.fromARGB(transparency, (byte)(wave - (x * 255)), (byte)(wave - (x * 255)), (byte)(x * 255));

			int catX = (int)((double)row / height * 40);
			int catY = (int)((double)col / width * 40);
			if ( 	catX == 0 || catX == 10 || catX == 20 || catX == 30 || catX == 40 
				 || catY == 0 || catY == 10 || catY == 20 || catY == 30 || catY == 40 ) {
				transparency = (byte)(255 - ( (double)catX/40*255 + ((double)catY/40*255 ) ) / 2);
				pixel = UInt32Utils.fromARGB(transparency, 0xFF, 0xFF, 0xFF);
			}

			//copy pixel into texture
			TextureUtils.SetPtrVal2D(data, pixel, row, col, width);
		}

		//vertical bars, red to green
		unsafe private void FillTexure2(uint* data, int row, int col, int width, int height)
		{
			//crate position in texture in the range [0..1]
			var x = 1; //(double)row / height;
			var y = (double)col / width;

			//make some waves in the range [0..255]
			var wave = (byte)(255 * ((Math.Sin(x * y * FWaveCountIn[FCurrentSlice] * 20) + 1) / 2.0));

			//a pixel is just a 32-bit unsigned int value
			byte transparency = wave; //(byte)(wave * (1-y));
			var pixel = UInt32Utils.fromARGB(transparency, (byte)(wave - (y * 255)), (byte)(y * 255), 0);
			
			int catX = (int)((double)row / height * 40);
			int catY = (int)((double)col / width * 40);
			if ( 	catX == 0 || catX == 10 || catX == 20 || catX == 30 || catX == 40 
				 || catY == 0 || catY == 10 || catY == 20 || catY == 30 || catY == 40 ) {
				transparency = (byte)(255 - ( (double)catX/40*255 + ((double)catY/40*255 ) ) / 2);
				pixel = UInt32Utils.fromARGB(transparency, 0xFF, 0xFF, 0xFF);
			}

			//copy pixel into texture
			TextureUtils.SetPtrVal2D(data, pixel, row, col, width);
		}

		#endregion FillTexture functions

		private VertexBuffer CreateQuadTexturedVertexBuffer(Device device) {
/*			float[,] aQuad = { 	{-1.0f,1.0f,10.0f,0.0f,0.0f,0.0f,0.0f},
								{-1.0f,-1.0f,10.0f,0.0f,1.0f,0.0f,1.0f},
								{1.0f,1.0f,10.0f,1.0f,0.0f,1.0f,0.0f},
								{1.0f,-1.0f,10.0f,1.0f,1.0f,1.0f,1.0f} };
*/

			float xy = 1.0f;
			float z = 0;
			float u = FTextureUChangeIn[0];
			float v = FTextureVChangeIn[0];

			TexturedVertex[] tv = new TexturedVertex[4];
			tv[0].Position.X = -xy; 			tv[0].Position.Y = xy; 			tv[0].Position.Z = z; 
			tv[0].TextureCoordinate.X = u; 		tv[0].TextureCoordinate.Y = v;
			tv[1].Position.X = -xy; 				tv[1].Position.Y = -xy; 			tv[1].Position.Z = z;
			tv[1].TextureCoordinate.X = u; 	tv[1].TextureCoordinate.Y = 1+v;
			tv[2].Position.X = xy; 			tv[2].Position.Y = xy; 		tv[2].Position.Z = z;
			tv[2].TextureCoordinate.X = 1+u; 		tv[2].TextureCoordinate.Y = v;
			tv[3].Position.X = xy /*- (float)(FBlendIn[0] / 2)*/; 				tv[3].Position.Y = -xy /*+ (float)(FBlendIn[0] / 2)*/; 		tv[3].Position.Z = z;
			tv[3].TextureCoordinate.X = 1+u; 	tv[3].TextureCoordinate.Y = 1+v;
			int tvsize = tv.GetLength(0) * Marshal.SizeOf( typeof( TexturedVertex ) );
			FLogger.Log(LogType.Debug, "tvsize = " + tvsize );
			//VertexBuffer vb = new VertexBuffer( typeof(TexturedVertex), 4, d, Usage.WriteOnly | Usage.Dynamic, TexturedVertex.Format, Pool.Default );
			//public VertexBuffer(Type typeVertexType, int numVerts, Device device, Usage usage, VertexFormats vertexFormat, Pool pool);
			VertexBuffer vb = new VertexBuffer(device, tvsize, Usage.WriteOnly | Usage.Dynamic, TexturedVertex.Format, Pool.Default);

			FillVertexBuffer( vb, tv, tvsize );

			//Mesh m = new Mesh(d, 2, 4, MeshFlags.DoNotClip, TexturedVertex.Format);
			//vb = m.VertexBuffer;
			/*
			DataStream vbds = vb.Lock(0, tvsize, LockFlags.None);
			IntPtr tvp = Marshal.UnsafeAddrOfPinnedArrayElement(tv, 0);
			//Marshal.Copy( tvp, 0, vbds.DataPointer, tvsize );
			for (int i = 0; i < tvsize; i++) {
				try {
					vbds.WriteByte(Marshal.ReadByte(tvp, i));
				} catch {}
			}
			//Unlock the vb before you can use it elsewhere
			vb.Unlock();
			*/
			
			/*
			ColoredVertex[] cv = new ColoredVertex[4];
			cv[0].Position.X = -xy; cv[0].Position.Y = xy; cv[0].Position.Z = z; 
			cv[0].Color = 0xFF00FF;
			cv[1].Position.X = xy; cv[1].Position.Y = xy; cv[1].Position.Z = z;
			cv[1].Color = 0xFF00FF;
			cv[2].Position.X = xy; cv[2].Position.Y = -xy; cv[2].Position.Z = z;
			cv[2].Color = 0xFF00FF;
			cv[3].Position.X = -xy; cv[3].Position.Y = -xy; cv[3].Position.Z = z;
			cv[3].Color = 0xFF00FF;
			int cvsize = tv.GetLength(0) * Marshal.SizeOf( typeof( TexturedVertex ) );				
			vb = new VertexBuffer(d, cvsize, Usage.WriteOnly, ColoredVertex.Format, Pool.Default);
			
			FillVertexBuffer( vb, cv, cvsize );

			d.SetStreamSource(0, vb, 0, Marshal.SizeOf( typeof( ColoredVertex ) ) );
			d.SetStreamSource(1, vb, 0, Marshal.SizeOf( typeof( ColoredVertex ) ) );
			*/

			return vb;
		}

		private void FillVertexBuffer<T>( VertexBuffer vertexBuffer, T[] vertexArray, int vertexarraySizeInBytes ) {
			DataStream vbds = vertexBuffer.Lock(0, vertexarraySizeInBytes, LockFlags.None);
			IntPtr tvp = Marshal.UnsafeAddrOfPinnedArrayElement(vertexArray, 0);
			//Marshal.Copy( tvp, 0, vbds.DataPointer, tvsize );
			for (int i = 0; i < vertexarraySizeInBytes; i++) {
				try {
					vbds.WriteByte(Marshal.ReadByte(tvp, i));
				} catch {}
			}
			//Unlock the vb before you can use it elsewhere
			vertexBuffer.Unlock();
		}
	}
}
